
#ifndef CPP_START_FUNCTION_TRAITS_H
#define CPP_START_FUNCTION_TRAITS_H

#include <functional>
#include <tuple>

//function_traits 获取普通函数、函数指针、std::function、函数对象以及成员函数的函数类型、返回类型、参数个数和参数的具体类型。

//转换为std::function和函数指针
template<typename T>
struct function_traits;

//普通函数
//返回类型Ret 参数列表Args
template<typename Ret, typename... Args>
struct function_traits<Ret(Args...)> {
public:
    enum {
        arity = sizeof...(Args)
    }; //enum 定义参数列表的个数arity常量
    //定义别名
    typedef Ret function_type(Args...);

    typedef Ret return_type;
    using std_function_type = std::function<function_type>;

    typedef Ret (*pointer)(Args...);

    template<size_t I>
    struct args {
        //static_assert(常量表达式，错误提示字符串) 静态断言，
        static_assert(I < arity, "index is out of range,index less than sizeof Args");

        //std::tuple_element 获取元素的类型
        using type = typename std::tuple_element<I, std::tuple<Args...>>::type;
    };
};

template<typename Ret, typename... Args>
struct function_traits<Ret(*)(Args...)> : function_traits<Ret(Args...)> {
};

template<typename Ret, typename... Args>
struct function_traits<std::function<Ret(Args...)>> : function_traits<Ret(Args...)> {
};

#define FUNCTION_TRAITS(...) \
    template <typename ReturnType,typename ClassType,typename... Args> \
    struct function_traits<ReturnType(ClassType::*)(Args...) __VA_ARGS__>:function_traits<ReturnType(Args...)>{}; \

FUNCTION_TRAITS()
FUNCTION_TRAITS(const)
FUNCTION_TRAITS(volatile)
FUNCTION_TRAITS(const volatile)

//函数对象
template<typename Callable>
//decltype  类型说明符 作用时选择并返回操作数的数据类型.
struct function_traits : function_traits<decltype(&Callable::operator())> {
};

// lambda表达式是一个匿名类，使用一个lambda表达式，实际就是使用一个匿名类的匿名对象，匿名类内部包含operator()调用符。
// operator()的返回类型、形参列表以及 函数体与lambda表达式相对应，因此可像函数调用一样使用lambda表达式。
//可通过function_traits获取lambda表达式的operator()类型，之后再将其转换为std::function

template<typename Function>
typename function_traits<Function>::std_function_type to_function(const Function &lambda) {
    return static_cast<typename function_traits<Function>::std_function_type>(lambda);
}

template<typename Function>
typename function_traits<Function>::std_function_type to_function(Function &&lamdba) {
    return static_cast<typename function_traits<Function>::std_function_type>(std::forward<Function>(lamdba));
}

template<typename Function>
typename function_traits<Function>::pointer to_function_pointer(const Function &lambda) {
    return static_cast<typename function_traits<Function>::pointer>(lambda);
}

#endif //CPP_START_FUNCTION_TRAITS_H
